struct tree {  int data;
               struct tree* left;
               struct tree* right; };

/*@ inductive reachable{L}(struct tree* t1, struct tree* t2) {
  case single: \forall struct tree* t; reachable(t,t);
  case transl: \forall struct tree* t1,*t2; reachable(t1->left,t2) ==> 
  reachable(t1,t2);
  case transr: \forall struct tree* t1, *t2; 
  reachable(t1->right,t2) ==> reachable(t1,t2);
}
*/

/*@ lemma reachable_trans_l:
  \forall struct tree *t1, *t2;
  reachable(t1,t2) && \valid(t2) ==> reachable(t1, t2->left);
*/

/*@ lemma reachable_trans_r:
  \forall struct tree *t1, *t2;
  reachable(t1,t2) && \valid(t2) ==> reachable(t1, t2->right);
*/

/*@ predicate well_formed{L}(struct tree* t) =
  \forall struct tree* t1; reachable(t,t1) ==>
  \valid(t1) || t1 == (struct tree*)0;
*/

/*@ predicate contains{L}(struct tree* t, int data) =
  \valid(t) &&
  \exists struct tree* t1; \valid(t1) && reachable(t,t1) && t1->data == data;
*/

/*@ inductive sorted_tree{L}(struct tree* t) {
  case empty{L}: sorted_tree((struct tree*)0);
  case full{L}: \forall struct tree* t; well_formed(t) &&
  \valid(t) && sorted_tree(t->left) && sorted_tree(t->right) &&
  (\forall struct tree* t1; reachable(t->left, t1) ==> t->data >= t1->data) &&
  (\forall struct tree* t1; reachable(t->right, t1) ==> t->data <= t1->data) ==>
  sorted_tree(t);
}
*/

/*@ lemma sorted_tree_projl:
  \forall struct tree* t, *t1;
  sorted_tree(t) && \valid(t) && \valid(t1) && reachable(t->left,t1) ==>
  t1->data <= t->data;
*/

/*@ lemma sorted_tree_projr:
  \forall struct tree* t, *t1;
  sorted_tree(t) && \valid(t) && \valid(t1) && reachable(t->right,t1) ==>
  t1->data >= t->data;
*/

/*@ lemma sorted_children:
  \forall struct tree* t;
  \valid(t) && sorted_tree(t) ==> sorted_tree(t->left) && sorted_tree(t->right);
*/

/* @ lemma reachable_choice{L}:
  \forall struct tree* t, *t1;
  \valid(t) && \valid(t1) && well_formed(t) && reachable(t,t1) ==>
  t == t1 || reachable(t->left, t1) || reachable(t->right,t1);
*/

/*@ lemma contains_choice{L}:
  \forall struct tree* t; \forall int data;
  \valid(t) && contains(t,data) ==>
  t->data == data || contains(t->left, data) || contains(t->right, data);
*/

/*@ 
  requires sorted_tree(t);
  requires well_formed(t);
  assigns \nothing;
  behavior find:
  assumes contains(t,key);
  ensures \valid(\result);
  ensures \result->data == key;
  behavior not_find:
  assumes !contains(t,key);
  ensures \result == (struct tree*)0;
  complete behaviors;
  disjoint behaviors;
 */
struct tree* search(int key, struct tree* t) {
  struct tree* current = t;
  /*@ 
    loop invariant reachable(t,current);
    loop invariant well_formed(current);
    loop invariant sorted_tree(current);
    loop invariant contains(t,key) <==> contains(current,key);
    loop assigns current;
  */
  while (current) {
    //@ assert \valid(current);
    if (current->data == key) return current;
    if (current->data < key)
      /*@ assert \forall struct tree* t1;
        \valid(t1) && reachable(current->left,t1) ==> t1->data < key; */
      //@ assert !contains(current->left,key);
      current = current->right;
    else
      //@ assert !contains(current->right,key);
      current = current -> left;
  }
  return current; }

/*
Local Variables:
compile-command: "frama-c-gui -wp -wp-rte binary_search_tree_proved.c"
End:
*/
